# -*- coding: utf-8 -*-

"""
Custom base target definition.
"""


__all__ = ["Target"]


import os
from abc import abstractmethod, abstractproperty

import luigi

from law.util import colored


class Target(luigi.target.Target):

    def __repr__(self):
        return "<{} at {}>".format(self.__class__.__name__, hex(id(self)))

    def colored_repr(self):
        return "{}({})".format(colored(self.__class__.__name__, "cyan"), hex(id(self)))

    def status_text(self, max_depth=0, color=True):
        if self.exists():
            text = "existent"
            _color = "green"
        else:
            text = "absent"
            _color = "red"

        return colored(text, _color, style="bright") if color else text

    @abstractmethod
    def exists(self):
        pass

    @abstractmethod
    def remove(self, silent=True):
        pass

    @abstractproperty
    def hash(self):
        pass



# -*- coding: utf-8 -*-

"""
Collections that wrap multiple targets.
"""


__all__ = ["TargetCollection", "SiblingFileCollection"]


import types

import six

from law.target.base import Target
from law.target.file import FileSystemTarget
from law.util import colored, flatten, create_hash


class TargetCollection(Target):

    def __init__(self, targets, threshold=1.0):
        if isinstance(targets, types.GeneratorType):
            targets = list(targets)
        elif not isinstance(targets, (list, tuple, dict)):
            raise TypeError("invalid targets, must be of type: list, tuple, dict")

        super(TargetCollection, self).__init__()

        # store targets and threshold
        self.targets = targets
        self.threshold = threshold

        # store flat targets for simplified iterations
        self._flat_targets = flatten(self.targets)

    def __repr__(self):
        return "<{}(len={}, threshold={}) at {}>".format(self.__class__.__name__, len(self),
            self.threshold, hex(id(self)))

    def colored_repr(self):
        return "{}(len={}, threshold={})".format(colored(self.__class__.__name__, "cyan"),
            colored(len(self), style="bright"), colored(self.threshold, style="bright"))

    def __len__(self):
        return len(self.targets)

    def __getitem__(self, key):
        return self.targets[key]

    def __iter__(self):
        raise TypeError("'{}' object is not iterable".format(self.__class__.__name__))

    @property
    def hash(self):
        target_hashes = "".join(target.hash for target in self._flat_targets)
        return create_hash(self.__class__.__name__ + target_hashes)

    def remove(self, silent=True):
        for target in self._flat_targets:
            target.remove(silent=silent)

    def _threshold(self):
        if self.threshold < 0:
            return 0
        elif self.threshold <= 1:
            return len(self) * self.threshold
        else:
            return min(len(self), self.threshold)

    def exists(self):
        threshold = self._threshold()

        # trivial case
        if threshold == 0:
            return True

        # simple counting with early stopping criteria for both success and fail
        n = 0
        for i, target in enumerate(self._flat_targets):
            if target.exists():
                n += 1
                if n >= threshold:
                    return True

            if n + (len(self) - i - 1) < threshold:
                return False

        return False

    def count(self, existing=True):
        # simple counting
        n = 0
        for target in self._flat_targets:
            if target.exists():
                n += 1

        return n if existing else len(self) - n

    def status_text(self, max_depth=0, color=True):
        count = self.count()
        exists = count >= self._threshold()

        if exists:
            text = "existent"
            _color = "green"
        else:
            text = "absent"
            _color = "red"

        text = colored(text, _color, style="bright") if color else text
        text += " ({}/{})".format(count, len(self))

        if max_depth > 0:
            if isinstance(self.targets, (list, tuple)):
                gen = enumerate(self.targets)
            else: # dict
                gen = six.iteritems(self.targets)

            for key, target in gen:
                text += "\n{}: ".format(key)

                if isinstance(target, TargetCollection):
                    text += "\n  ".join(target.status_text(max_depth - 1, color=color).split("\n"))
                elif isinstance(target, Target):
                    text += "{} ({})".format(target.status_text(color=color), target.colored_repr())

        return text


class SiblingFileCollection(TargetCollection):

    def __init__(self, *args, **kwargs):
        super(SiblingFileCollection, self).__init__(*args, **kwargs)

        # check if all targets are file system targets or nested SiblingFileCollection's
        # (it's the user's responsibility to pass targets that are really in the same directory)
        for target in self._flat_targets:
            if not isinstance(target, (FileSystemTarget, SiblingFileCollection)):
                raise TypeError("SiblingFileCollection's only wrap FileSystemTarget's and "
                    "other SiblingFileCollection's, got {}".format(target.__class__))

        # find the first target and store its directory
        first_target = self._flat_targets[0]
        if isinstance(first_target, FileSystemTarget):
            self.dir = first_target.parent
        else: # SiblingFileCollection
            self.dir = first_target.dir

    def exists(self, basenames=None):
        threshold = self._threshold()

        # check the dir
        if not self.dir.exists():
            return False

        # trivial case
        if threshold == 0:
            return True

        # get the basenames of all elements of the directory
        if basenames is None:
            basenames = self.dir.listdir()

        # simple counting with early stopping criteria for both success and fail
        n = 0
        for i, target in enumerate(self._flat_targets):
            if isinstance(target, FileSystemTarget):
                if target.basename in basenames:
                    n += 1
            else: # SiblingFileCollection
                if target.exists(basenames=basenames):
                    n += 1

            # we might be done here
            if n >= threshold:
                return True
            elif n + (len(self) - i - 1) < threshold:
                return False

        return False

    def count(self, existing=True, basenames=None):
        # trivial case when the contained directory does not exist
        if not self.dir.exists():
            return 0

        # get the basenames of all elements of the directory
        if basenames is None:
            basenames = self.dir.listdir()

        # simple counting
        n = 0
        for target in self._flat_targets:
            if isinstance(target, FileSystemTarget):
                if target.basename in basenames:
                    n += 1
            else: # SiblingFileCollection
                if target.exists(basenames=basenames):
                    n += 1

        return n if existing else len(self) - n



# -*- coding: utf-8 -*-

"""
Bash sandbox implementation.
"""


__all__ = ["BashSandbox"]


import os
from collections import OrderedDict
from fnmatch import fnmatch
from subprocess import PIPE, STDOUT

import luigi
import six

import law
from law.sandbox.base import Sandbox
from law.config import Config
from law.util import make_list, tmp_file, interruptable_popen


class BashSandbox(Sandbox):

    sandbox_type = "bash"

    # env cache per init script
    _envs = {}

    @property
    def script(self):
        return os.path.expandvars(os.path.expanduser(self.name))

    @property
    def env(self):
        # strategy: create a tempfile, let python dump its full env in a subprocess. and load the
        # env file again afterwards
        script = self.script
        if script not in self._envs:
            with tmp_file() as tmp:
                tmp_path = os.path.realpath(tmp[1])

                cmd = "bash -l -c 'source \"{0}\"; python -c \"" \
                    "import os,pickle;pickle.dump(os.environ,open(\\\"{1}\\\",\\\"w\\\"))\"'"
                cmd = cmd.format(script, tmp_path)

                returncode, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
                    stdout=PIPE, stderr=STDOUT)
                if returncode != 0:
                    raise Exception("bash sandbox env loading failed: " + str(out))

                with open(tmp_path, "r") as f:
                    env = six.moves.cPickle.load(f)

            # add env variables defined in the config
            env.update(self.get_config_env())

            # add env variables defined by the task
            env.update(self.get_task_env())

            # cache
            self._envs[script] = env

        return self._envs[script]

    def cmd(self, proxy_cmd):
        cfg = Config.instance()

        # environment variables to set
        env = OrderedDict()

        # sandboxing variables
        env["LAW_SANDBOX"] = self.key
        env["LAW_SANDBOX_SWITCHED"] = "1"

        # add env variables defined in the config and by the task
        env.update(self.get_config_env())
        env.update(self.get_task_env())

        # handle scheduling within the container
        ls_flag = "--local-scheduler"
        if self.force_local_scheduler() and ls_flag not in proxy_cmd:
            proxy_cmd.append(ls_flag)
        if ls_flag not in proxy_cmd:
            if getattr(self.task, "_worker_id", None):
                env["LAW_SANDBOX_WORKER_ID"] = self.task._worker_id
            if getattr(self.task, "_worker_task", None):
                env["LAW_SANDBOX_WORKER_TASK"] = self.task._worker_task

        # build commands to add env variables
        pre_cmds = []
        for tpl in env.items():
            pre_cmds.append("export {}=\"{}\"".format(*tpl))

        # build the final command
        cmd = "bash -l -c 'source \"{script}\"; {pre_cmd}; {proxy_cmd}'".format(
            proxy_cmd=" ".join(proxy_cmd), pre_cmd="; ".join(pre_cmds), script=self.script)

        return cmd

    def get_config_env(self):
        return super(BashSandbox, self).get_config_env("bash_env_" + self.script, "bash_env")

    def get_task_env(self):
        return super(BashSandbox, self).get_task_env("get_bash_env")



# -*- coding: utf-8 -*-

"""
Base definition of a minimalistic job manager.
"""


__all__ = ["BaseJobManager", "BaseJobFile"]


import os
import time
import shutil
import tempfile
import fnmatch
from abc import ABCMeta, abstractmethod

import six

from law.util import colored


@six.add_metaclass(ABCMeta)
class BaseJobManager(object):

    PENDING = "pending"
    RUNNING = "running"
    RETRY = "retry"
    FINISHED = "finished"
    FAILED = "failed"
    UNKNOWN = "unknown"

    status_names = [PENDING, RUNNING, RETRY, FINISHED, FAILED, UNKNOWN]

    # color styles per status when job count increases / decreases
    status_diff_styles = {
        PENDING: ({}, {"color": "green"}),
        RUNNING: ({"color": "green"}, {}),
        RETRY: ({"color": "red"}, {}),
        FINISHED: ({"color": "green"}, {}),
        FAILED: ({"color": "red", "style": "bright"}, {}),
        UNKNOWN: ({"color": "red", "style": "bright"}, {}),
    }

    @abstractmethod
    def submit(self, *args, **kwargs):
        pass

    @abstractmethod
    def submit_batch(self, *args, **kwargs):
        pass

    @abstractmethod
    def cancel(self, *args, **kwargs):
        pass

    @abstractmethod
    def cancel_batch(self, *args, **kwargs):
        pass

    @abstractmethod
    def cleanup(self, *args, **kwargs):
        pass

    @abstractmethod
    def cleanup_batch(self, *args, **kwargs):
        pass

    @abstractmethod
    def query(self, *args, **kwargs):
        pass

    @abstractmethod
    def query_batch(self, *args, **kwargs):
        pass

    @classmethod
    def job_status_dict(cls, job_id=None, status=None, code=None, error=None):
        return dict(job_id=job_id, status=status, code=code, error=error)

    @classmethod
    def status_line(cls, counts, last_counts=None, skip=None, timestamp=True, align=False,
        color=False):
        status_names = self.status_names
        if skip:
            status_names = [name for name in status_names if name not in skip]

        # check last counts
        if last_counts and len(last_counts) != len(status_names):
            raise Exception("{} last status counts expected, got {}".format(len(status_names),
                len(last_counts)))

        # check current counts
        if len(counts) != len(status_names):
            raise Exception("{} status counts expected, got {}".format(len(status_names),
                len(counts)))

        # calculate differences
        if last_counts:
            diffs = tuple(n - m for n, m in zip(counts, last_counts))

        # number formatting
        if isinstance(align, bool) or not isinstance(align, six.integer_types):
            align = 4 if align else 0
        count_fmt = "%d" if not align else "%%%sd" % align
        diff_fmt = "%+d" if not align else "%%+%sd" % align

        # build the status line
        line = ""
        if timestamp:
            line += "{}: ".format(time.strftime("%H:%M:%S"))
        line += "all: " + count_fmt % (sum(counts),)
        for i, (status, count) in enumerate(zip(status_names, counts)):
            count = count_fmt % count
            if color:
                count = colored(count, style="bright")
            line += ", {}: {}".format(status, count)

            if last_counts:
                diff = diff_fmt % diffs[i]
                if color:
                    diff = colored(diff, **cls.status_diff_styles[status][diffs[i] < 0])
                line += " ({})".format(diff)

        return line


@six.add_metaclass(ABCMeta)
class BaseJobFile(object):

    config_attrs = []

    class Config(dict):

        def __getattr__(self, attr):
            return self.__getitem__(attr)

        def __setattr__(self, attr, value):
            self.__setitem__(attr, value)

    def __init__(self, tmp_dir=None):
        super(BaseJobFile, self).__init__()

        self.tmp_dir = tmp_dir if tmp_dir is not None else tempfile.mkdtemp()

    def __del__(self):
        self.cleanup()

    def __call__(self, *args, **kwargs):
        return self.create(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.cleanup()

    def cleanup(self):
        if isinstance(self.tmp_dir, six.string_types) and os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def provide_input(self, src, postfix, render_data=None):
        basename = os.path.basename(src)
        dst = os.path.join(self.tmp_dir, self.postfix_file(basename, postfix))
        if render_data:
            self.render_file(src, dst, render_data)
        else:
            shutil.copy2(src, dst)
        return dst

    def get_config(self, kwargs):
        cfg = self.Config()
        for attr in self.config_attrs:
            cfg[attr] = kwargs.get(attr, getattr(self, attr))
        return cfg

    @abstractmethod
    def create(self, postfix=None, render=None, **kwargs):
        pass

    @classmethod
    def postfix_file(cls, path, postfix):
        if postfix:
            if isinstance(postfix, six.string_types):
                _postfix = postfix
            else:
                basename = os.path.basename(path)
                for pattern, _postfix in six.iteritems(postfix):
                    if fnmatch.fnmatch(basename, pattern):
                        break
                else:
                    _postfix = ""
            path = "{1}{0}{2}".format(_postfix, *os.path.splitext(path))
        return path

    @classmethod
    def render_file(cls, src, dst, render_data):
        with open(src, "r") as f:
            lines = f.readlines()

        basename = os.path.basename(src)
        for pattern, variables in six.iteritems(render_data):
            if fnmatch.fnmatch(basename, pattern):
                for key, value in six.iteritems(variables):
                    lines = [cls.render_line(line, key, value) for line in lines]

        with open(dest, "w") as f:
            for line in lines:
                f.write(line)

    @classmethod
    def render_line(cls, line, key, value):
        return line.replace("{{" + key + "}}", value)

    @classmethod
    def create_line(cls, key, value=None, indent=0):
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""
DCache file system and targets.
"""


__all__ = ["DCacheFileSystem", "DCacheFileTarget", "DCacheDirectoryTarget"]


import os

from law.config import Config
from law.target.remote import RemoteFileSystem, RemoteTarget, RemoteTarget, RemoteFileTarget, \
    RemoteDirectoryTarget


class DCacheFileSystem(RemoteFileSystem):

    default_instance = None

    def __init__(self, config=None, base=None, bases=None, **kwargs):
        # prepare the gfal options
        # resolution order: config, base+bases, default dCache section
        if not config and not base:
            config = Config.instance().get("target", "default_dcache")
        if config and Config.instance().has_section(config):
            base = Config.instance().get_default(config, "base")
            bases = {}
            prefix = "base_"
            for key, value in Config.instance().items(config):
                if key.startswith(prefix):
                    bases[key[len(prefix):]] = value
        if base is None:
            raise Exception("invalid arguments, set either config, base or the "
                "target.default_dcache option in your law config")

        # default configs
        kwargs.setdefault("retries", 1)
        kwargs.setdefault("retry_delay", 5)
        kwargs.setdefault("transfer_config", {"checksum_check": False})
        kwargs.setdefault("validate_copy", False)
        kwargs.setdefault("cache_config", {})
        kwargs.setdefault("atomic_contexts", True)
        kwargs.setdefault("permissions", False)

        super(DCacheFileSystem, self).__init__(base, bases, **kwargs)


# try to set the default fs instance
try:
    DCacheFileSystem.default_instance = DCacheFileSystem()
except:
    pass


class DCacheTarget(RemoteTarget):

    def __init__(self, path, fs=DCacheFileSystem.default_instance):
        """ __init__(path, fs=DCacheFileSystem.default_instance)
        """
        RemoteTarget.__init__(self, path, fs)


class DCacheFileTarget(DCacheTarget, RemoteFileTarget):

    pass


class DCacheDirectoryTarget(DCacheTarget, RemoteDirectoryTarget):

    pass


DCacheTarget.file_class = DCacheFileTarget
DCacheTarget.directory_class = DCacheDirectoryTarget



# -*- coding: utf-8 -*-

"""
Main law command line interface.
"""


import os
import sys
from importlib import import_module
from argparse import ArgumentParser


progs = ["run", "db", "config", "software", "completion"]
forward_progs = ["run"]


def main():
    # setup the main parser and sub parsers
    parser = ArgumentParser(prog="law", description="law command line tool")
    sub_parsers = parser.add_subparsers(help="subcommands", dest="command")

    # setup all progs
    mods = {}
    for prog in progs:
        mods[prog] = import_module("law.scripts." + prog)
        mods[prog].setup_parser(sub_parsers)

    # parse args and dispatch execution
    if sys.argv[1] in forward_progs:
        args = parser.parse_args(sys.argv[1:3])
    else:
        args = parser.parse_args()

    mods[args.command].execute(args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
Simple tasks that build a tree.

The first task (FetchLoremIpsum) will download six different versions of a "lorem Ipsum" text. The
next task (CountChars) determines and saves the frequency of every character. After that, the (json)
count files are merged (MergeCounts). The last task (ShowFrequencies) illustrates the "measured"
frequencies and publishes the result to the scheduler.
"""


import json
import time
import random
from collections import defaultdict

from six.moves import urllib
import luigi
import law


URL = "http://www.loremipsum.de/downloads/version%i.txt"

luigi.namespace("example.loremipsum")


class LoremIpsumBase(law.Task):

    version = luigi.IntParameter(description="the file version in the range of 1 to 6")
    slow = luigi.BoolParameter(description="before running, wait between 5 and 15 seconds")

    def wait(self):
        if self.slow:
            time.sleep(random.randint(5, 15))


class FetchLoremIpsum(LoremIpsumBase):

    def __init__(self, *args, **kwargs):
        super(FetchLoremIpsum, self).__init__(*args, **kwargs)

        if not (1 <= self.version <= 6):
            raise ValueError("version must be in the range of 1 to 6")

    def output(self):
        return law.LocalFileTarget("data/loremipsum/data_%i.txt" % self.version)

    def run(self):
        self.wait()

        with self.output().localize() as t:
            urllib.request.urlretrieve(URL % self.version, t.path)

        # this is the same as
        #
        #   output = self.output()
        #   output.parent.touch()
        #   url = URL % self.version
        #   urllib.urlretrieve(URL % self.version, output.path)
        #   output.chmod(0o0660)
        #
        # but only for local targets, whereas the above code also works
        # for remote targets!


class CountChars(LoremIpsumBase):

    def requires(self):
        return FetchLoremIpsum.req(self)

    def output(self):
        return law.LocalFileTarget("data/loremipsum/chars_%i.json" % self.version)

    def run(self):
        self.wait()

        # open("r") works for local _and_ remote targets
        with self.input().open("r") as f:
            content = f.read()

        content = content.replace("\n", " ").lower()
        counts = {}
        for c in "abcdefghijklmnopqrstuvwxyz":
            counts[c] = content.count(c)

        # open("w") also works for local _and_ remote targets
        # but keep in mind that chmod and parent.touch must be invoked
        # manually (which is not necessary in this example)
        with self.output().open("w") as f:
            json.dump(counts, f, indent=4)


class MergeCounts(LoremIpsumBase):

    # this task has no version
    version = None

    def requires(self):
        return [CountChars.req(self, version=i) for i in range(1, 7)]

    def output(self):
        return law.LocalFileTarget("data/loremipsum/chars_merged.json")

    def run(self):
        self.wait()

        merged_counts = defaultdict(int)

        for target in self.input():
            with target.open("r") as f:
                counts = json.load(f)
            for c, count in counts.items():
                merged_counts[c] += count

        with self.output().open("w") as f:
            json.dump(merged_counts, f, indent=4)


class ShowFrequencies(LoremIpsumBase):

    version = None

    def requires(self):
        return MergeCounts.req(self)

    def output(self):
        return law.LocalFileTarget("data/loremipsum/char_frequencies.txt")

    def run(self):
        self.wait()

        with self.input().open("r") as f:
            counts = json.load(f)

        # normalize and convert to frequency (%)
        count_sum = sum(counts.values())
        counts = {char: int(round(100 * count / count_sum)) for char, count in counts.items()}

        # sort (char, freq) descending by count
        counts = sorted(counts.items(), key=lambda tpl: -tpl[1])

        # prepare output content
        content = "\n".join("%s: %s %i%%" % (char, "|" * count, count) for char, count in counts)

        # write to output
        with self.output().open("w") as f:
            f.write(content)

        # also send as message to publisher
        self.publish_message(content)


luigi.namespace()



# -*- coding: utf-8 -*-

"""
Helpers for working with the Worldwide LHC Computing Grid.
"""


__all__ = ["get_voms_proxy_user", "get_voms_proxy_lifetime", "get_voms_proxy_vo",
           "renew_voms_proxy", "delegate_voms_proxy_glite", "get_ce_endpoint"]


import os
import time
import re
import subprocess
import uuid
import json

import six

from law.util import interruptable_popen, tmp_file, create_hash


def _voms_proxy_info(args=None):
    cmd = ["voms-proxy-info"] + (args or [])
    code, out, err = interruptable_popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if code != 0:
        raise Exception("voms-proxy-info failed: {}".format(err))
    return code, out, err


def get_voms_proxy_user():
    out = _voms_proxy_info(["--identity"])[1].strip()
    try:
        return re.match(".*\/CN\=([^\/]+).*", out.strip()).group(1)
    except:
        raise Exception("no valid identity found in voms proxy: {}".format(out))


def get_voms_proxy_lifetime():
    out = _voms_proxy_info(["--timeleft"])[1].strip()
    try:
        return int(out)
    except:
        raise Exception("no valid lifetime found in voms proxy: {}".format(out))


def get_voms_proxy_vo():
    return _voms_proxy_info(["--vo"])[1].strip()


def renew_voms_proxy(vo, passwd, valid="196:00"):
    with tmp_file() as (_, tmp):
        with open(tmp, "w") as f:
            f.write(passwd)

        cmd = "cat '{}' | voms-proxy-init -voms '{}' --valid '{}'".format(tmp, vo, valid)
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if code != 0:
            raise Exception("proxy renewal failed: {}".format(out))


def delegate_voms_proxy_glite(endpoint, stdout=None, stderr=None, cache=True):
    # get the proxy file
    proxy_file = os.environ.get("X509_USER_PROXY", "/tmp/x509up_u%i" % os.getuid())
    if not os.path.exists(proxy_file):
        raise Exception("proxy file '{}' does not exist".format(proxy_file))

    if cache:
        if isinstance(cache, six.string_types):
            cache_file = cache
        else:
            cache_file = proxy_file + "_delegation_cache.json"

        def remove_cache():
            try:
                if os.path.exists(cache_file):
                    os.remove(cache_file)
            except OSError:
                pass

        # create the hash of the proxy file content
        with open(proxy_file, "r") as f:
            proxy_hash = create_hash(f.read())

        # already delegated?
        cache_data = {}
        if os.path.exists(cache_file):
            with open(cache_file, "r") as f:
                try:
                    cache_data = json.load(f)
                except:
                    remove_cache()

        # is the hash up-to-date?
        if cache_data.get("hash") != proxy_hash:
            remove_cache()
            cache_data = {}

        # proxy already delegated to that endpoint?
        elif endpoint in cache_data.get("ids", []):
            return str(cache_data["ids"][endpoint])

    # do the actual delegation
    delegation_id = uuid.uuid4().hex
    cmd = ["glite-ce-delegate-proxy", "-e", endpoint, delegation_id]
    code = interruptable_popen(cmd, stdout=stdout, stderr=stderr)[0]
    if code != 0:
        raise Exception("glite proxy delegation to endpoint {} failed".format(endpoint))

    if cache:
        # write the id back to the delegation file
        cache_data["hash"] = proxy_hash
        cache_data.setdefault("ids", {})[endpoint] = delegation_id
        with open(cache_file, "w") as f:
            json.dump(cache_data, f, indent=4)

    return delegation_id


def get_ce_endpoint(ce):
    return ce.split("/", 1)[0]



# -*- coding: utf-8 -*-

"""
HTCondor job implemention.
"""


__all__ = ["HTCondorJobManager", "HTCondorJobFile"]


import os
import sys
import time
import re
import random
import subprocess
import getpass
from multiprocessing.pool import ThreadPool

import six

from law.job.base import BaseJobManager, BaseJobFile
from law.util import interruptable_popen, iter_chunks, make_list, multi_match


class HTCondorJobManager(BaseJobManager):

    submission_job_id_cre = re.compile("^(\d+)\sjob\(s\)\ssubmitted\sto\scluster\s(\d+)\.$")
    status_header_cre = re.compile("^\s*ID\s+.+$")
    status_line_cre = re.compile("^(\d+\.\d+)\s+[^\s]+\s+[^\s]+\s+[^\s]+\s+[^\s]+\s+([UIRXCHE])\s+.*$")
    history_cluster_id_cre = re.compile("^ClusterId\s=\s(\d+)$")
    history_process_id_cre = re.compile("^ProcId\s=\s(\d+)$")
    history_code_cre = re.compile("^ExitCode\s=\s(-?\d+)$")
    history_reason_cre = re.compile("^RemoveReason\s=\s\"(.*)\"$")

    def __init__(self, pool=None, scheduler=None, threads=1):
        super(HTCondorJobManager, self).__init__()

        self.pool = pool
        self.scheduler = scheduler
        self.threads = threads

    def cleanup(self, *args, **kwargs):
        raise NotImplementedError("HTCondorJobManager.cleanup is not implemented")

    def cleanup_batch(self, *args, **kwargs):
        raise NotImplementedError("HTCondorJobManager.cleanup_batch is not implemented")

    def submit(self, job_file, pool=None, scheduler=None, retry_delay=5, silent=False):
        # default arguments
        pool = pool or self.pool
        scheduler = scheduler or self.scheduler

        # build the command
        cmd = ["condor_submit"]
        if pool:
            cmd += ["-pool", pool]
        if scheduler:
            cmd += ["-name", scheduler]
        cmd += [os.path.basename(job_file)]

        # define the actual submission in a loop to simplify retries
        while True:
            # run the command
            code, out, err = interruptable_popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                cwd=os.path.dirname(job_file))

            # get the job id(s)
            if code == 0:
                last_line = out.strip().split("\n")[-1].strip()
                m = self.submission_job_id_cre.match(last_line)
                if m:
                    job_ids = ["{}.{}".format(m.group(2), i) for i in range(int(m.group(1)))]
                else:
                    code = 1
                    err = "cannot parse job id(s) from output:\n{}".format(out)

            # retry or done?
            if code == 0:
                return job_ids
            else:
                if retries > 0:
                    retries -= 1
                    time.sleep(retry_delay)
                    continue
                elif silent:
                    return None
                else:
                    raise Exception("submission of job '{}' failed:\n{}".format(job_file, err))

    def submit_batch(self, job_files, pool=None, scheduler=None, retries=0, retry_delay=5,
        silent=False, threads=None):
        # default arguments
        threads = threads or self.threads

        # threaded processing
        kwargs = dict(pool=pool, scheduler=scheduler, retries=retries, retry_delay=retry_delay,
            silent=silent)
        pool = ThreadPool(max(threads, 1))
        results = [pool.apply_async(self.submit, (job_file,), kwargs) \
                   for job_file in job_files]
        pool.close()
        pool.join()

        # store return values or errors
        outputs = []
        for res in results:
            try:
                outputs.append(res.get())
            except Exception as e:
                outputs.append(e)

        return outputs

    def cancel(self, job_id, pool=None, scheduler=None, silent=False):
        # default arguments
        pool = pool or self.pool
        scheduler = scheduler or self.scheduler

        # build the command
        cmd = ["condor_rm"]
        if pool:
            cmd += ["-pool", pool]
        if scheduler:
            cmd += ["-name", scheduler]
        cmd += make_list(job_id)

        # run it
        code, out, err = interruptable_popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # check success
        if code != 0 and not silent:
            raise Exception("cancellation of job(s) '{}' failed:\n{}".format(job_id, err))

    def cancel_batch(self, job_ids, pool=None, scheduler=None, silent=False, threads=None,
        chunk_size=20):
        # default arguments
        pool = pool or self.pool
        scheduler = scheduler or self.scheduler
        threads = threads or self.threads

        # threaded processing
        kwargs = dict(pool=pool, scheduler=scheduler, silent=silent)
        pool = ThreadPool(max(threads, 1))
        results = [pool.apply_async(self.cancel, (job_id_chunk,), kwargs) \
                   for job_id_chunk in iter_chunks(job_ids, chunk_size)]
        pool.close()
        pool.join()

        # store errors
        errors = []
        for res in results:
            try:
                res.get()
            except Exception as e:
                errors.append(e)

        return errors

    def query(self, job_id, pool=None, scheduler=None, user=None, silent=False):
        # default arguments
        pool = pool or self.pool
        scheduler = scheduler or self.scheduler

        multi = isinstance(job_id, (list, tuple))

        # query the condor queue
        cmd = ["condor_q", "-nobatch"]
        if pool:
            cmd += ["-pool", pool]
        if scheduler:
            cmd += ["-name", scheduler]
        cmd += make_list(job_id)
        code, out, err = interruptable_popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # handle errors
        if code != 0:
            if silent:
                return None
            else:
                raise Exception("queue query of job(s) '{}' failed:\n{}".format(job_id, err))

        # parse the output and extract the status per job
        query_data = self.parse_queue_output(out)

        # find missing jobs, and query the condor history for the exit code
        missing_ids = [_job_id for _job_id in make_list(job_id) if _job_id not in query_data]
        if missing_ids:
            cmd = ["condor_history", user or getpass.getuser()]
            cmd += ["--long", "--attributes", "ClusterId,ProcId,ExitCode,RemoveReason"]
            if pool:
                cmd += ["-pool", pool]
            if scheduler:
                cmd += ["-name", scheduler]
            code, out, err = interruptable_popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            # handle errors
            if code != 0:
                if silent:
                    return None
                else:
                    raise Exception("history query of job(s) '{}' failed:\n{}".format(job_id, err))

            # parse the output and update query data
            query_data.update(self.parse_history_output(out, job_ids=missing_ids))

        # compare to the requested job ids and perform some checks
        for _job_id in make_list(job_id):
            if _job_id not in query_data:
                if not multi:
                    if silent:
                        return None
                    else:
                        raise Exception("job(s) '{}' not found in query response".format(job_id))
                else:
                    query_data[_job_id] = self.job_status_dict(job_id=_job_id,
                        error="job not found in query response")

        return query_data if multi else query_data[job_id]

    def query_batch(self, job_ids, pool=None, scheduler=None, user=None, silent=False, threads=None,
        chunk_size=20):
        # default arguments
        threads = threads or self.threads

        # threaded processing
        kwargs = dict(pool=pool, scheduler=scheduler, user=user, silent=silent)
        pool = ThreadPool(max(threads, 1))
        results = [pool.apply_async(self.query, (job_id_chunk,), kwargs) \
                   for job_id_chunk in iter_chunks(job_ids, chunk_size)]
        pool.close()
        pool.join()

        # store status data per job id
        query_data, errors = {}, []
        for res in results:
            try:
                query_data.update(res.get())
            except Exception as e:
                errors.append(e)

        return query_data, errors

    @classmethod
    def parse_queue_output(cls, out):
        query_data = {}

        header = None
        for line in out.strip().split("\n"):
            if cls.status_header_cre.match(line):
                header = line
            elif header:
                m = cls.status_line_cre.match(line)
                if m:
                    job_id = m.group(1)
                    status_flag = m.group(2)

                    # map the status
                    status = cls.map_status(status_flag)

                    # save the result
                    query_data[job_id] = cls.job_status_dict(job_id=job_id, status=status)
                else:
                    break

        return query_data

    @classmethod
    def parse_history_output(cls, out, job_ids=None):
        # build blocks per job, i.e. output lines that are separated by empty lines
        blocks = []
        for line in out.strip().split("\n"):
            line = line.strip()
            if not blocks or not line:
                blocks.append([])
            if not line:
                continue
            blocks[-1].append(line)

        # helper to extract job data from a block
        attrs = ["cluster_id", "code", "process_id", "reason"]
        cres = [getattr(cls, "history_%s_cre" % (attr,)) for attr in attrs]
        def parse_block(block):
            data = {}
            for line in block:
                for attr, cre in six.moves.zip(attrs, cres):
                    if attr not in data:
                        m = cre.match(line)
                        if m:
                            data[attr] = m.group(1)
                            break
            return data

        # extract job information per block
        query_data = {}
        for block in blocks:
            data = parse_block(sorted(block))

            if "cluster_id" not in data and "process_id" not in data:
                continue

            # interpret data
            job_id = "{cluster_id}.{process_id}".format(**data)
            code = data.get("code") and int(data["code"])
            error = data.get("reason")
            if error and code in (0, None):
                code = 1
            status = cls.FINISHED if code == 0 else cls.FAILED

            # store it
            query_data[job_id] = cls.job_status_dict(job_id=job_id, status=status, code=code,
                error=error)

        return query_data

    @classmethod
    def map_status(cls, status_flag):
        # see http://pages.cs.wisc.edu/~adesmet/status.html
        if status_flag in ("U", "I", "H"):
            return cls.PENDING
        elif status_flag in ("R",):
            return cls.RUNNING
        elif status_flag in ("C",):
            return cls.FINISHED
        elif status_flag in ("E",):
            return cls.FAILED
        else:
            return cls.UNKNOWN


class HTCondorJobFile(BaseJobFile):

    config_attrs = ["file_name", "universe", "executable", "arguments", "input_files",
        "output_files", "stdout", "stderr", "log", "notification", "custom_content"]

    def __init__(self, file_name="job", universe="vanilla", executable=None, arguments=None,
        input_files=None, output_files=None, stdout="stdout.txt", stderr="stderr.txt",
        log="log.txt", notification="Never", custom_content=None, tmp_dir=None):
        super(HTCondorJobFile, self).__init__(tmp_dir=tmp_dir)

        self.file_name = file_name
        self.universe = universe
        self.executable = executable
        self.arguments = arguments
        self.input_files = input_files or []
        self.output_files = output_files or []
        self.stdout = stdout
        self.stderr = stderr
        self.log = log
        self.notification = notification
        self.custom_content = custom_content

    def create(self, postfix=None, render_data=None, **kwargs):
        # merge kwargs and instance attributes
        c = self.get_config(kwargs)

        # some sanity checks
        if not c.file_name:
            raise ValueError("file_name must not be empty")
        elif not c.universe:
            raise ValueError("universe must not be empty")
        elif not c.executable:
            raise ValueError("executable must not be empty")

        # prepare paths
        job_file = self.postfix_file(os.path.join(self.tmp_dir, c.file_name), postfix)
        c.input_files = map(os.path.abspath, c.input_files)
        executable_is_file = c.executable in map(os.path.basename, c.input_files)

        # prepare input files
        c.input_files = [self.provide_input(path, postfix, render_data) for path in c.input_files]
        if executable_is_file:
            c.executable = self.postfix_file(os.path.basename(c.executable), postfix)

        # output files
        c.output_files = [self.postfix_file(path, postfix) for path in c.output_files]
        c.stdout = c.stdout and self.postfix_file(c.stdout, postfix)
        c.stderr = c.stdout and self.postfix_file(c.stderr, postfix)
        c.log = c.log and self.postfix_file(c.log, postfix)

        # job file content
        content = []
        content.append(("Universe", c.universe))
        content.append(("Executable", c.executable))
        if c.input_files:
            content.append(("transfer_input_files", c.input_files))
        if c.output_files:
            content.append(("transfer_output_files", c.output_files))
        if c.stdout:
            content.append(("Output", c.stdout))
        if c.stderr:
            content.append(("Error", c.stderr))
        if c.log:
            content.append(("Log", c.log))
        if c.notification:
            content.append(("Notification", c.notification))

        # add custom content
        if c.custom_content:
            content += c.custom_content

        # finally arguments and queuing statements
        if c.arguments:
            for _arguments in make_list(c.arguments):
                content.append(("Arguments", _arguments))
                content.append("Queue")

        # write the job file
        with open(job_file, "w") as f:
            for obj in content:
                f.write(self.create_line(*make_list(obj)) + "\n")

        return job_file

    @classmethod
    def create_line(cls, key, value=None):
        if isinstance(value, (list, tuple)):
            value = ",".join(str(v) for v in value)
        if value is None:
            return str(key)
        else:
            return "{} = {}".format(key, value)



# -*- coding: utf-8 -*-

"""
"law db" command line tool.
"""


import os
import sys
from importlib import import_module

import luigi
import six

from law.task.base import Task
from law.config import Config
from law.util import multi_match


def setup_parser(sub_parsers):
    parser = sub_parsers.add_parser("db", prog="law db", description="law db file updater")

    parser.add_argument("--modules", "-m", nargs="+", help="additional modules to traverse")
    parser.add_argument("--remove", "-r", action="store_true", help="just remove the db file")
    parser.add_argument("--verbose", "-v", action="store_true", help="verbose output")


def execute(args):
    # just remove the db file?
    if args.remove:
        db_file = Config.instance().get("core", "db_file")
        if os.path.exists(db_file):
            os.remove(db_file)
        return

    # get modules to lookup
    lookup = [m.strip() for m in Config.instance().keys("modules")]
    if args.modules:
        lookup += args.modules

    print("loading tasks from {} modules".format(len(lookup)))

    # loop through modules, import everything to load tasks
    for modid in lookup:
        if not modid:
            continue

        if args.verbose:
            print("loading module '{}'".format(modid))

        try:
            mod = import_module(modid)
        except ImportError as e:
            continue

        if args.verbose:
            print("loaded module '{}'".format(modid))

    if args.verbose:
        print("")

    # determine tasks to write into the db file
    seen_families = []
    task_classes = []
    lookup = [Task]
    while lookup:
        cls = lookup.pop(0)
        lookup.extend(cls.__subclasses__())

        if cls.task_family in seen_families:
            continue
        seen_families.append(cls.task_family)

        skip = cls.exclude_db or not six.callable(getattr(cls, "run", None)) \
            or getattr(cls.run, "__isabstractmethod__", False)
        if not skip:
            task_classes.append(cls)

            if args.verbose:
                print("add task '{}'".format(cls.task_family))
        else:
            if args.verbose:
                print("skip task '{}'".format(cls.task_family))

    def dbline(cls, default_namespace=None):
        # determine parameters
        params = []
        for attr in dir(cls):
            member = getattr(cls, attr)
            if isinstance(member, luigi.Parameter):
                exclude = getattr(cls, "exclude_params_db", set())
                if not multi_match(attr, exclude, any):
                    params.append(attr.replace("_", "-"))

        # build and return the line
        # format: "module_id:task_family:param param ..."
        return "{}:{}:{}".format(cls.__module__, cls.task_family, " ".join(params))

    # write the db file
    db_file = Config.instance().get("core", "db_file")
    if not os.path.exists(os.path.dirname(db_file)):
        os.makedirs(os.path.dirname(db_file))

    with open(db_file, "w") as f:
        for cls in task_classes:
            f.write(dbline(cls) + "\n")

    print("written {} task(s) to db file {}".format(len(task_classes), db_file))



# -*- coding: utf-8 -*-

"""
Workflow implementation based on gLite job submission. See
https://wiki.italiangrid.it/twiki/bin/view/CREAM/UserGuide.
"""


__all__ = ["GLiteWorkflow"]


import os
import sys
import math
import random
import base64
from collections import OrderedDict, defaultdict
from abc import abstractmethod

import luigi
import six

from law.workflow.base import Workflow, WorkflowProxy
from law.parameter import CSVParameter
from law.decorator import log
from law.util import law_base, iter_chunks
from law.contrib.job.glite import GLiteJobManager, GLiteJobFile
from law.contrib.util.wlcg import delegate_voms_proxy_glite, get_ce_endpoint


class GLiteSubmissionData(OrderedDict):

    dummy_job_id = "dummy_job_id"

    def __init__(self, **kwargs):
        super(GLiteSubmissionData, self).__init__()

        self["jobs"] = kwargs.get("jobs", {})
        self["tasks_per_job"] = kwargs.get("tasks_per_job", 1)

    @classmethod
    def job_data(cls, job_id=dummy_job_id, branches=None):
        return dict(job_id=job_id, branches=branches or [])

    @property
    def jobs(self):
        return self["jobs"]

    @property
    def tasks_per_job(self):
        return self["tasks_per_job"]


class GLiteStatusData(OrderedDict):

    dummy_job_id = "dummy_job_id"

    def __init__(self, **kwargs):
        super(GLiteStatusData, self).__init__()

        self["jobs"] = kwargs.get("jobs", {})

    @classmethod
    def job_data(cls, job_id=dummy_job_id, status=None, code=None, error=None):
        return GLiteJobManager.job_status_dict(job_id, status, code, error)

    @property
    def jobs(self):
        return self["jobs"]


class GLiteWorkflowProxy(WorkflowProxy):

    workflow_type = "glite"

    submission_data_cls = GLiteSubmissionData

    status_data_cls = GLiteStatusData

    def __init__(self, *args, **kwargs):
        super(GLiteWorkflowProxy, self).__init__(*args, **kwargs)

        self.job_file = None
        self.job_manager = GLiteJobManager()
        self.delegation_ids = None
        self.submission_data = self.submission_data_cls(tasks_per_job=self.task.tasks_per_job)
        self.skipped_job_nums = None
        self.last_counts = len(self.job_manager.status_names) * (0,)
        self.retry_counts = defaultdict(int)

    def requires(self):
        task = self.task
        reqs = OrderedDict()

        # add upstream requirements when not cancelling or cleaning
        if not task.cancel_jobs and not task.cleanup_jobs:
            reqs.update(super(GLiteWorkflowProxy, self).requires())

        return reqs

    def output(self):
        task = self.task

        # get the directory where the control outputs are stored
        out_dir = task.glite_output_directory()

        # define outputs
        outputs = OrderedDict()
        postfix = task.glite_output_postfix()

        # a file containing the submission data, i.e. job ids etc
        submission_file = "submission{}.json".format(postfix)
        outputs["submission"] = out_dir.child(submission_file, type="f")

        # a file containing status data when the jobs are done
        if not task.no_poll and not task.cancel_jobs and not task.cleanup_jobs:
            status_file = "status{}.json".format(postfix)
            outputs["status"] = out_dir.child(status_file, type="f")

        # when not cancelling or cleaning, update with actual outputs
        if not task.cancel_jobs and not task.cleanup_jobs:
            outputs.update(super(GLiteWorkflowProxy, self).output())

        return outputs

    @log
    def run(self):
        task = self.task
        outputs = self.output()

        # read submission data and reset some values
        submitted = outputs["submission"].exists()
        if submitted:
            self.submission_data.update(outputs["submission"].load(formatter="json"))
            task.tasks_per_job = self.submission_data.tasks_per_job

        # when the branch outputs, i.e. the "collection" exists, just create dummy control outputs
        if outputs["collection"].exists():
            self.touch_control_outputs()

        # cancel jobs?
        elif task.cancel_jobs:
            if submitted:
                self.cancel()

        # cleanup jobs?
        elif task.cleanup_jobs:
            if submitted:
                self.cleanup()

        # submit and/or wait while polling
        else:
            outputs["submission"].parent.touch()

            # at this point, when the status file exists, it is considered outdated
            if "status" in outputs:
                outputs["status"].remove()

            try:
                self.job_file = GLiteJobFile()

                # submit
                if not submitted:
                    self.submit()

                # start status polling when a) no_poll is not set, or b) the jobs were already
                # submitted so that failed jobs are resubmitted after a single polling iteration
                if not task.no_poll or submitted:
                    self.poll()

            finally:
                # finally, cleanup the job file
                if self.job_file:
                    self.job_file.cleanup()

    def create_job_file(self, job_num, branches):
        task = self.task
        config = {}

        thisdir = os.path.dirname(os.path.abspath(__file__))
        def rel_file(*paths):
            return os.path.normpath(os.path.join(thisdir, *paths))

        # the file postfix is pythonic range made from branches, e.g. [0, 1, 2] -> "_0To3"
        _postfix = "_{}To{}".format(branches[0], branches[-1] + 1)
        postfix = lambda path: self.job_file.postfix_file(path, _postfix)
        config["postfix"] = {"*": _postfix}

        # executable
        config["executable"] = rel_file("wrapper.sh")

        # input files
        config["input_files"] = [rel_file("wrapper.sh"), law_base("job", "job.sh")]

        # render variables
        config["render"] = defaultdict(dict)
        config["render"]["*"]["job_file"] = postfix("job.sh")

        # add the bootstrap file
        bootstrap_file = task.glite_bootstrap_file()
        config["input_files"].append(bootstrap_file)
        config["render"]["*"]["bootstrap_file"] = postfix(os.path.basename(bootstrap_file))

        # output files
        config["output_files"] = []

        # log file
        if task.transfer_logs:
            log_file = "stdall.txt"
            config["stdout"] = log_file
            config["stderr"] = log_file
            config["output_files"].append(log_file)
            config["render"]["*"]["log_file"] = postfix(log_file)
        else:
            config["stdout"] = None
            config["stderr"] = None

        # output uri
        config["output_uri"] = task.glite_output_uri()

        # job script arguments
        task_params = task.as_branch(branches[0]).cli_args(exclude={"branch"})
        job_args = [
            task.__class__.__module__,
            task.task_family,
            base64.b64encode(" ".join(task_params)).replace("=", "_"),
            str(branches[0]),
            str(branches[-1] + 1),
            "no"
        ]
        config["render"]["wrapper.sh"]["job_args"] = " ".join(job_args)

        # task hook
        config = task.glite_job_config(config)

        return self.job_file(**config)

    def cancel(self):
        task = self.task

        # get job ids from submission data
        job_ids = [d["job_id"] for d in self.submission_data.jobs.values() \
                   if d["job_id"] not in (self.submission_data.dummy_job_id, None)]
        if not job_ids:
            return

        # cancel jobs
        task.publish_message("going to cancel {} jobs".format(len(job_ids)))
        errors = self.job_manager.cancel_batch(job_ids, threads=task.threads)

        # print errors
        if errors:
            print("{} errors occured while cancelling {} jobs:".format(len(errors), len(job_ids)))
            print("\n".join(str(e) for e in errors))

    def cleanup(self):
        task = self.task

        # get job ids from submission data
        job_ids = [d["job_id"] for d in self.submission_data.jobs.values() \
                   if d["job_id"] not in (self.submission_data.dummy_job_id, None)]
        if not job_ids:
            return

        # cleanup jobs
        task.publish_message("going to cleanup {} jobs".format(len(job_ids)))
        errors = self.job_manager.cleanup_batch(job_ids, threads=task.threads)

        # print errors
        if errors:
            print("{} errors occured while cleanup {} jobs:".format(len(errors), len(job_ids)))
            print("\n".join(str(e) for e in errors))

    def submit(self, job_map=None):
        task = self.task

        # delegate the voms proxy to all endpoints
        if self.delegation_ids is None and callable(task.glite_delegate_proxy):
            self.delegation_ids = []
            for ce in self.ce:
                endpoint = get_ce_endpoint(ce)
                self.delegation_ids.append(task.glite_delegate_proxy(endpoint))

        # map branch numbers to job numbers, chunk by tasks_per_job
        if not job_map:
            branch_chunks = list(iter_chunks(task.branch_map, task.tasks_per_job))
            job_map = dict((i + 1, branches) for i, branches in enumerate(branch_chunks))

        # when only_missing is True, a job can be skipped when all its tasks are completed
        check_skip = False
        if task.only_missing and self.skipped_job_nums is None:
            self.skipped_job_nums = []
            check_skip = True

        # create job files for each chunk
        job_data = OrderedDict()
        for job_num, branches in six.iteritems(job_map):
            if check_skip and all(task.as_branch(b).complete() for b in branches):
                self.skipped_job_nums.append(job_num)
                self.submission_data.jobs[job_num] = self.submission_data_cls.job_data(
                    branches=branches)
                continue

            # create and store the job file
            job_data[job_num] = (branches, self.create_job_file(job_num, branches))

        # actual submission
        job_files = [job_file for _, job_file in six.itervalues(job_data)]
        task.publish_message("going to submit {} jobs to {}".format(len(job_files), task.ce))
        job_ids = self.job_manager.submit_batch(job_files, ce=task.ce,
            delegation_id=self.delegation_ids, retries=3, threads=task.threads)

        # store submission data
        errors = []
        for job_num, job_id in six.moves.zip(job_data, jobs):
            if isinstance(job_id, Exception):
                errors.append((job_num, job_id))
                job_id = None
            self.submission_data.jobs[job_num] = self.submission_data_cls.job_data(job_id=job_id,
                branches=job_data[job_num][0])

        # write the submission data to the output file
        self.output()["submission"].dump(self.submission_data, formatter="json")

        # raise exceptions or log
        if errors:
            raise Exception("{} error(s) occured during submission:\n    {}".format(len(errors),
                "\n    ".join("job {}: {}".format(*tpl) for tpl in errors)))
        else:
            task.publish_message("submitted {} jobs to {}".format(len(job_files), task.ce))

    def poll(self):
        task = self.task
        outputs = self.output()

        # submission stats
        n_jobs = float(len(self.submission_data.jobs))
        n_finished_min = task.acceptance * n_jobs if task.acceptance <= 1 else task.acceptance
        n_failed_max = task.tolerance * n_jobs if task.tolerance <= 1 else task.tolerance
        max_polls = int(math.ceil((task.walltime * 3600.) / (task.interval * 60.)))
        n_poll_fails = 0

        # bookkeeping dicts to avoid querying the status of finished jobs
        # note: unfinished_jobs holds submission data, finished_jobs holds status data
        unfinsihed_jobs = {}
        finished_jobs = {}

        # fill dicts from submission data, taking into account skipped jobs
        for job_num, data in six.iteritems(self.submission_data.jobs):
            if self.skipped_job_nums and job_num in self.skipped_job_nums:
                finished_jobs[job_num] = self.status_data_cls.job_data(
                    status=self.job_manager.FINISHED, code=0)
            else:
                unfinished_jobs[job_num] = data.copy()
                unfinishedJobData[jobNum] = tpl

        # use maximum number of polls for looping
        for i in six.moves.range(max_polls):
            # sleep
            if i > 0:
                sleep(task.interval * 60)

            # query job states
            job_ids = [data["job_id"] for data in unfinished_jobs]
            states, errors = self.job_manager.query_batch(job_ids, threads=task.threads)
            if errors:
                print("{} error(s) occured during status query:\n    {}".format(len(errors),
                    "\n    ".join(str(e) for e in errors)))

                n_poll_fails += 1
                if n_poll_fails > task.max_poll_fails:
                    raise Exception("max_poll_fails exceeded")
                else:
                    continue
            else:
                n_poll_fails = 0

            # store jobs per status, remove finished ones from unfinished_jobs
            pending_jobs = {}
            running_jobs = {}
            failed_jobs = {}
            unknown_jobs = {}
            for job_num, data in six.iteritems(states):
                if data["status"] == self.job_manager.PENDING:
                    pending_jobs[job_num] = data
                elif data["status"] == self.job_manager.RUNNING:
                    running_jobs[job_num] = data
                elif data["status"] == self.job_manager.FINISHED:
                    finished_jobs[job_num] = data
                    unfinished_jobs.pop(job_num)
                elif data["status"] in (self.job_manager.FAILED, self.job_manager.RETRY):
                    failed_jobs[job_num] = data
                else:
                    unknown_jobs[job_num] = data

            # counts
            n_pending = len(pending_jobs)
            n_running = len(running_jobs)
            n_finished = len(finished_jobs)
            n_failed = len(failed_jobs)
            n_unknown = len(unknown_jobs)

            # determine jobs that failed and might be resubmitted
            retry_jobs = {}
            if n_failed and task.retries > 0:
                for job_num in failed_jobs:
                    if self.retry_counts[job_num] < task.retries:
                        retry_jobs[job_num] = self.submission_data.jobs[job_num]["branches"]
                        self.retry_counts[job_num] += 1
            n_retry = len(retry_jobs)
            n_failed -= n_retry

            # log the status line
            counts = (n_pending, n_running, n_retry, n_finished, n_failed, n_unknown)
            status_line = self.job_manager.status_line(counts, self.last_counts, color=True)
            task.publish_message(status_line)
            self.last_counts = counts

            # log failed jobs
            if len(failed_jobs) > 0:
                print("reasons for failed jobs in task {}:".format(task.task_id))
                tmpl = "    {}: {}, {status}, {code}, {error}"
                for job_num, data in six.iteritems(failed_jobs):
                    job_id = self.submission_data.jobs[job_num]["job_id"]
                    print(tmpl.format(job_num, job_id, **data))

            # infer the overall status
            finished = n_finished >= n_finished_min
            failed = n_failed > n_failed_max
            unreachable = n_jobs - n_failed < n_finished_min
            if finished:
                # write status output
                if "status" in outputs:
                    status_data = self.status_data_cls()
                    status_data.jobs.update(finished_jobs)
                    status_data.jobs.update(states)
                    outputs["status"].dump(status_data, formatter="json")
                break
            elif failed:
                failed_nums = [job_num for job_num in failed_jobs if job_num not in retry_jobs]
                raise Exception("tolerance exceeded for jobs {}".format(failed_nums))
            elif unreachable:
                raise Exception("acceptance of {} unreachable, total: {}, failed: {}".format(
                    n_finished_min, n_jobs, n_failed))

            # automatic resubmission
            if n_retry:
                self.skipped_job_nums = []
                self.submit_jobs(retry_jobs)

                # update the unfinished so the next iteration is aware of the new job ids
                for job_num in retry_jobs:
                    unfinished_jobs[job_num]["job_id"] = self.submission_data.jobs[job_num]["job_id"]

            # break when no polling is desired
            # we can get to this point when there was already a submission and the no_poll
            # parameter was set so that only failed jobs are submitted again
            if task.no_poll:
                break
        else:
            # walltime exceeded
            raise Exception("walltime exceeded")

    def touch_control_outputs(self):
        task = self.task

        # create the parent directory
        outputs = self.output()
        outputs["submission"].parent.touch()

        # get all branch indexes and chunk them by tasks_per_job
        branch_chunks = list(iter_chunks(task.branch_map, task.tasks_per_job))

        # submission output
        if not outputs["submission"].exists():
            submission_data = self.submission_data.copy()
            # set dummy submission data
            submission_data.jobs.clear()
            for i, branches in enumerate(branch_chunks):
                job_num = i + 1
                submission_data.jobs[job_num] = self.submission_data_cls.job_data(branches=branches)
            outputs["submission"].dump(submission_data, formatter="json")

        # status output
        if "status" in outputs and not outputs["status"].exists():
            status_data = self.status_data_cls()
            # set dummy status data
            for i, branches in enumerate(branch_chunks):
                job_num = i + 1
                status_data.jobs[job_num] = self.status_data_cls.job_data(
                    status=self.job_manager.FINISHED, code=0)
            outputs["status"].dump(status_data, formatter="json")


class GLiteWorkflow(Workflow):

    exclude_db = True

    workflow_proxy_cls = GLiteWorkflowProxy

    ce = CSVParameter(default=[], significant=False, description="target computing element(s)")
    retries = luigi.IntParameter(default=5, significant=False, description="number of automatic "
        "resubmission attempts per job, default: 5")
    tasks_per_job = luigi.IntParameter(default=1, significant=False, description="number of tasks "
        "to be processed by one job, default: 1")
    only_missing = luigi.BoolParameter(significant=False, description="skip tasks that are "
        "considered complete")
    no_poll = luigi.BoolParameter(significant=False, description="just submit, do not initiate "
        "status polling after submission")
    threads = luigi.IntParameter(default=4, significant=False, description="number of threads to "
        "use for (re)submission and status queries, default: 4")
    interval = luigi.FloatParameter(default=3, significant=False, description="time between status "
        "polls in minutes, default: 3")
    walltime = luigi.FloatParameter(default=48, significant=False, description="maximum wall time "
        "in hours, default: 48")
    max_poll_fails = luigi.IntParameter(default=5, significant=False, description="maximum number "
        "of consecutive errors during polling, default: 5")
    cancel_jobs = luigi.BoolParameter(default=False, description="cancel all submitted jobs, no "
        "new submission")
    cleanup_jobs = luigi.BoolParameter(default=False, description="cleanup all submitted jobs, no "
        "new submission")
    transfer_logs = luigi.BoolParameter(significant=False, description="transfer job logs to the "
        "output directory")

    exclude_params_branch = {"ce", "retries", "tasks_per_job", "only_missing", "no_poll", "threads",
        "interval", "walltime", "max_poll_fails", "cancel_jobs", "cleanup_jobs", "transfer_logs"}

    @abstractmethod
    def glite_output_directory(self):
        return None

    @abstractmethod
    def glite_bootstrap_file(self):
        pass

    def glite_output_postfix(self):
        # TODO: use start/end branch?
        return ""

    def glite_output_uri(self):
        return self.glite_output_directory().url()

    def glite_delegate_proxy(self, endpoint):
        return delegate_voms_proxy_glite(endpoint, stdout=sys.stdout, stderr=sys.stderr,
            cache=True)

    def glite_job_config(self, config):
        return config


from law.contrib import provide
provide(locals())
